"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.upsertByName = exports.where = exports.cache = exports.put = exports.evaluateFilter = exports.parseCacheInstructions = exports.clearCacheDb = exports.FILTER_ARG = void 0;
const dotprop_1 = __importDefault(require("dotprop"));
const shared_1 = require("../shared");
exports.FILTER_ARG = 'filters';
const postDotPropFns = Object.keys(Object.getOwnPropertyDescriptors(Array.prototype));
const filterFunctions = {
    map: (dataBeingReduced, ...dotPropFilters) => {
        if (dotPropFilters.length > 1) {
            throw new Error('mapping with dot prop filters should not have more than one argument');
        }
        return dataBeingReduced.map((obj) => {
            const mapRet = (0, dotprop_1.default)(obj, dotPropFilters[0]);
            return mapRet;
        });
    }
};
const isValidFnCall = (fnCall) => {
    if (fnCall === null)
        return false;
    if (typeof fnCall.fnName !== 'string' || fnCall.fnName.length === 0)
        return false;
    if (!Array.isArray(fnCall.args))
        return false;
    return true;
};
const isValidCustomCall = (fnCall) => {
    if (!isValidFnCall(fnCall)) {
        return false;
    }
    if (!filterFunctions[fnCall.fnName])
        return false;
    return fnCall.isCustomCall === true;
};
const isValidArrayFnCall = (fnCall) => {
    if (!isValidFnCall(fnCall)) {
        return false;
    }
    if (!Array.prototype[fnCall.fnName])
        return false;
    if (fnCall.isCustomCall)
        return false;
    return true;
};
const isNum = (val) => /^\d+$/.test(val);
const isValidDotprop = (obj) => {
    if (typeof obj !== 'object')
        return false;
    if (typeof obj.string !== 'string' || obj.string.length === 0)
        return false;
    if ((typeof obj.validWith !== 'object') && obj.validWith !== '?')
        return false;
    return true;
};
const parseDotprop = (str) => {
    return {
        string: str.trim(),
        validWith: '?'
    };
};
const parseCustomCall = (str) => {
    const fnNames = Object.keys(filterFunctions);
    const regExpStr = `(${fnNames.join('|')})\\(([a-zA-Z0-9\.\-\_\/]+)(?:,\s*([a-zA-Z0-9\.\-\_\/]+)|)(?:,\s*([a-zA-Z0-9\.\-\_\/]+)|)(?:,\s*([a-zA-Z0-9\.\-\_\/]+)|)\\)`;
    const regexp = new RegExp(regExpStr);
    const matched = str.match(regexp);
    if (matched === null || !matched.filter)
        return null;
    const result = matched
        .filter((elem) => elem !== undefined);
    const result2 = result.map((elem) => {
        return isNum(elem) ? parseInt(elem) : elem;
    });
    result2.shift();
    const fnName1 = result2.shift();
    const parsedCall = {
        fnName: fnName1,
        args: result2,
        isCustomCall: true
    };
    return parsedCall;
};
const simpleParseFnCall = (str) => {
    const regExpStr = `Array\\.(${postDotPropFns.join('|')})\\(([0-9]+)(?:,\s*([0-9]+)|)\\)`;
    const regexp = new RegExp(regExpStr);
    const matched = str.match(regexp);
    if (matched === null || !matched.filter)
        return null;
    const result = matched
        .filter((elem) => elem !== undefined);
    const result2 = result.map((elem) => {
        return isNum(elem) ? parseInt(elem) : elem;
    });
    result2.shift();
    const fnName1 = result2.shift();
    if (typeof fnName1 !== 'string' || !Array.prototype[fnName1])
        throw new Error(`fnName should be in Array.prototype.`);
    const parsedCall = {
        fnName: fnName1,
        args: result2,
        isCustomCall: false
    };
    return parsedCall;
};
const parseFnOrDotprop = (str) => {
    const parsedCustomCall = parseCustomCall(str);
    if (isValidCustomCall(parsedCustomCall)) {
        return parsedCustomCall;
    }
    const parsedCall = simpleParseFnCall(str);
    if (isValidArrayFnCall(parsedCall)) {
        return parsedCall;
    }
    const parsedDotprop = parseDotprop(str);
    if (!isValidDotprop(parsedDotprop)) {
        throw new Error(`Query stage is neither a dotprop entry nor a custom filter function nor a  simple array-based function call: "${str}"`);
    }
    return parsedDotprop;
};
const parseQuery = (splitted) => {
    const cmds = splitted.filter(elem => !!elem);
    return cmds.map(parseFnOrDotprop);
};
const runFilter = (data, stages) => {
    return stages.reduce((accum, curr) => {
        let final;
        if (isValidDotprop(curr)) {
            final = (0, dotprop_1.default)(accum, curr.string);
        }
        else if (isValidCustomCall(curr)) {
            const currFnName = curr.fnName;
            if (typeof currFnName !== 'string')
                throw new Error(`Property fnName should have been string of length in ${curr}`);
            if (currFnName.length < 1)
                throw new Error(`Property fnName should have been string of length in ${curr}`);
            const callable = filterFunctions[currFnName];
            if (typeof callable !== 'function')
                throw new Error(`Tried to run custom filter ${currFnName}, but it was not present or not a function. Function-call data: ${curr}`);
            final = callable(accum, ...curr.args);
        }
        else {
            if (!isValidArrayFnCall(curr))
                throw new Error(`An unfit custom filter entry was found: ${curr}`);
            final = (Array.prototype[curr.fnName]).call(accum, ...curr.args);
        }
        return final;
    }, data);
};
const BRACKETS = ['{{', '}}'];
const requireValidCacheInstructions = (str) => {
    const reasons = [];
    const result = [true, reasons];
    if (typeof str !== 'string') {
        reasons.push('invalid type: ' + str + ' is ' + typeof str);
    }
    if (reasons.length)
        throw new Error(reasons.join('\n'));
    return result;
};
const clearCacheDb = async () => {
    const db = await (0, shared_1.get)('db');
    return db.cache.toCollection().delete();
};
exports.clearCacheDb = clearCacheDb;
const chooseCommands = (explicitCmds) => {
    if (typeof explicitCmds === 'string') {
        return explicitCmds.split(' ').filter(x => x !== '' && x !== ' ');
    }
    else {
        return [];
    }
};
const cacheInstructionsToQuery = (inner, defaultQuery) => {
    if (inner === null)
        return { commands: [], names: [], [exports.FILTER_ARG]: null };
    const split = inner.split('`');
    switch (split.length) {
        case 1:
            const cmds = chooseCommands(split[0]);
            return {
                commands: cmds,
                names: [],
                [exports.FILTER_ARG]: [defaultQuery]
            };
        case 2:
            const cmds2 = chooseCommands(split[0]);
            const names2 = chooseCommands(split[1]);
            return {
                commands: cmds2,
                names: names2,
                [exports.FILTER_ARG]: [defaultQuery]
            };
        case 3:
            const cmds3 = chooseCommands(split[0]);
            const names3 = chooseCommands(split[1]);
            const jq3 = split[2].split(' ');
            return {
                commands: cmds3,
                names: names3,
                [exports.FILTER_ARG]: jq3 ? jq3 : [defaultQuery]
            };
        default:
            throw new Error('invalid cache instructions (uncaught)' + inner);
    }
};
const numberlessDeref = /\*/;
const numberDeref = /[a-z]*[0-9]*\*/;
const extractDerefSections_ = (str) => {
    return {
        cli: '', brackets: null, bracketed: null, prefix: null
    };
};
const extractPointerSections_ = (str) => {
    return {
        cli: '', brackets: null, bracketed: null, numPrefix: null, namePrefix: null
    };
};
const extractBracketedSections_ = (str) => {
    let splitTwo;
    const splitOne = str.split(BRACKETS[0]);
    if (splitOne.length === 1) {
        return { cli: str, brackets: null, bracketed: null };
    }
    const [leftOuter, rightOfBrackOne] = splitOne;
    const lastBrax = rightOfBrackOne.lastIndexOf(BRACKETS[1]);
    const startLast = lastBrax + BRACKETS[1].length;
    splitTwo = [
        rightOfBrackOne.substr(0, lastBrax),
        rightOfBrackOne.substr(startLast)
    ];
    const [inner, rightOuter] = splitTwo;
    return {
        cli: `${leftOuter} ${rightOuter}`,
        brackets: inner ?? null,
        bracketed: inner
            ? `${BRACKETS[0]}${inner}${BRACKETS[1]}`
            : null
    };
};
const filterResult = (result, targets, indexName) => {
    if (!targets.length)
        return true;
    if (targets === '*') {
        return true;
    }
    const ret = targets.filter((queryCmd, idx) => {
        const test = result[indexName][idx];
        if (test === undefined || queryCmd === undefined) {
            return false;
        }
        const tested = test.startsWith(queryCmd);
        return tested;
    });
    return ret.length === targets.length;
};
const queryArrayIndex = async (targets, indexName) => {
    const db = await (0, shared_1.get)('db');
    const q = db.cache;
    let res;
    if (targets === '*') {
        res = await q.toArray();
    }
    else {
        const first = targets.shift();
        res = await q.where(indexName)
            .startsWith(first)
            .and((possResult) => {
            return filterResult(possResult, [first, ...targets], indexName);
        })
            .toArray();
    }
    res.sort((a, b) => {
        if (a.createdAt > b.createdAt)
            return 1;
        else if (a.createdAt < b.createdAt)
            return -1;
        return 0;
    });
    return res;
};
const filterOnArrayIndex = async (arr, targets, targetName) => {
    return arr.filter(possResult => filterResult(possResult, targets, targetName));
};
const simpleFilter = (lowerFilter, res1) => {
    if (lowerFilter === 'first' || lowerFilter === 'f') {
        return res1[0];
    }
    else if (lowerFilter === 'last' || lowerFilter === 'l') {
        return res1[res1.length - 1];
    }
    return null;
};
const interpretQueryAsLodash = async (filter, res1) => {
    if (filter.length !== 1)
        return null;
    const lowerFilter = filter[0].toLowerCase();
    return simpleFilter(lowerFilter, res1);
};
const interpretQuery = async (query) => {
    const { commands = [], names = [], [exports.FILTER_ARG]: filter = null } = query;
    let res1 = [];
    if (commands.length) {
        res1 = await queryArrayIndex(commands, 'commands');
        if (names.length) {
            res1 = await filterOnArrayIndex(res1, names, 'names');
        }
    }
    else if (names.length) {
        res1 = await queryArrayIndex(names, 'names');
    }
    if (res1.length === 0)
        return res1;
    if (filter !== null) {
        const lodashRes = await interpretQueryAsLodash(filter, res1);
        if (lodashRes !== null) {
            return lodashRes;
        }
        else {
            return (0, exports.evaluateFilter)(res1, filter);
        }
    }
    return res1;
};
const parseCacheInstructions = async (str, defaultFilter = null) => {
    const { cli, brackets, bracketed } = extractBracketedSections_(str);
    requireValidCacheInstructions(str);
    const query = cacheInstructionsToQuery(brackets, defaultFilter);
    const res = await interpretQuery(query);
    const strPatch = typeof res === 'object' ? JSON.stringify(res) : res;
    return str.replace(bracketed, strPatch);
};
exports.parseCacheInstructions = parseCacheInstructions;
const evaluateFilter = async (obj, query) => {
    if (!Array.isArray(query))
        return obj;
    if (query.length === 1 && query[0] === '.')
        return obj;
    const stages = parseQuery(query);
    const result = runFilter(obj, stages);
    if (result === undefined)
        throw new Error(`Results of the filter "${query}" were undefined`);
    return JSON.parse(JSON.stringify(result));
};
exports.evaluateFilter = evaluateFilter;
const firstLines = (entry) => {
    if (!entry)
        return '(falsy entry)';
    try {
        return JSON.stringify(entry, null, 2).split('\n').slice(0, 5).join('\n');
    }
    catch (e) {
        console.error('Could not show json sample!');
        return '';
    }
};
const put = async (entry) => {
    let filtered;
    const db = await (0, shared_1.get)('db');
    try {
        const { commands, names, [exports.FILTER_ARG]: _jq, value: valueIn } = entry;
        const value = await (0, exports.evaluateFilter)(valueIn, _jq);
        const createdAt = Date.now();
        const props = { commands, names, value, createdAt };
        filtered = Object
            .entries(props)
            .reduce((accum, [key, val]) => {
            if (val !== undefined && val !== null) {
                return { ...accum, [key]: val };
            }
            return accum;
        }, {});
        return db.cache.put(filtered);
    }
    catch (e) {
        const message = `Could not put new entry into uidb.cache.
${firstLines(entry)} ...

internal query data (if we made it that far): 
${firstLines(filtered)}

Db error:
${e.message}
`;
        console.error(message);
    }
};
exports.put = put;
const cache = async (commands, data, names = [], _jq) => {
    (0, exports.put)({ commands, names, value: data, [exports.FILTER_ARG]: _jq });
};
exports.cache = cache;
const anyIsNull = (c) => {
    return c !== '*' && c.filter(c1 => c1 === null).length > 0;
};
const where = async (query) => {
    if (anyIsNull(query.commands || []))
        throw new Error('Null disallowed in commands[] query argument');
    const { [exports.FILTER_ARG]: _jq } = query;
    let rawResult = [];
    if (typeof query.id === 'number' && query.id !== -1) {
        rawResult = await (await (0, shared_1.get)('db')).cache.where({ id: query.id }).toArray();
        return (0, exports.evaluateFilter)(rawResult, _jq || null);
    }
    else {
        return interpretQuery({
            ...query,
            commands: query.commands ?? '*',
            names: query.names ?? '*',
            [exports.FILTER_ARG]: query[exports.FILTER_ARG] || null
        });
    }
};
exports.where = where;
const upsertByName = async (unknownDexie, tableName, name) => {
    const dexieDb = unknownDexie;
    if (dexieDb[tableName] === undefined)
        throw new Error('table does not exist; ' + tableName);
    const found = await dexieDb[tableName].where({ name }).first();
    if (found)
        return found;
    await dexieDb[tableName].add({ name });
    return dexieDb[tableName].where({ name }).first();
};
exports.upsertByName = upsertByName;
//# sourceMappingURL=store.js.map